#include "lisp.h"
#include "memory.h"
#include "neopixel.h"

#include <cstdlib>
#include <cstring>

#define MAX_SYMBOL_LEN  24

static char      SymbolTable[SYMBOL_TABLE_SIZE];
static cell      CellTable[SYMBOL_TABLE_SIZE]; // TODO(kyle): consider adding
                                               // cell-table specific param.


static void
Abort()
{
    neopxl::stop();
    abort();
}


symbol_t
intern_symbol(const char *s)
{
    char    *cursor = SymbolTable;

    if (!(std::strncmp(s, cursor+1, MAX_SYMBOL_LEN) == 0)) {
        cursor += cursor[0];
    }

    // cursor has reached the end of the symbol table, which means
    // we are out of room.
    if (cursor >= SymbolTable+SYMBOL_TABLE_SIZE) {
        Abort();
    }

    // cursor has reached the end of the symbol list, which means it
    // should be added to the symbol table.
    if (cursor[0] == 0) {
        cursor[0] = static_cast<uint8_t>(strnlen(s, MAX_SYMBOL_LEN));
        memcpy(cursor+1, s, static_cast<size_t>(cursor[0]));
        cursor[cursor[0]+1] = 0;
    }

    // return the symbol pointed to by the cursor.
    return (symbol_t)(cursor);
}


static cell *
FindEmptyCell()
{
    auto cursor = CellTable;

    while (cursor < (CellTable + SYMBOL_TABLE_SIZE)) {
        if (cursor->type == 0) {
            return cursor;
        }
    }

    return nullptr;
}



symbol_t
make_symbol(const char *s)
{
    cell  *cell = FindEmptyCell();
    if (cell == nullptr) {
        Abort();
    }
    
    cell->type = TYPE_SYMBOL;
    cell->sym = intern_symbol(s);

    return cell->sym;
}

